const Buffer = require('safe-buffer').Buffer;
const encoding = require("@walletconnect/encoding");
const utils = require("@walletconnect/utils");
const crypto = require('./crypto/fallback/index');

function getRandomValues(array) {
  for (var i = 0, l = array.length; i < l; i++) {
    array[i] = Math.floor(Math.random() * 256);
  }
  return array;
}
const MAX_BYTES = 65536;
const MAX_UINT32 = 4294967295;
window.Buffer = Buffer;
function randomBytes(size) {
  if (size > MAX_UINT32) throw new RangeError('requested too many random bytes')

  var bytes = Buffer.allocUnsafe(size)
  if (size > 0) {  // getRandomValues fails on IE if size == 0
    if (size > MAX_BYTES) { // this is the max bytes crypto.getRandomValues
      // can do at once see https://developer.mozilla.org/en-US/docs/Web/API/window.crypto.getRandomValues
      for (var generated = 0; generated < size; generated += MAX_BYTES) {
        // buffer.slice automatically checks if the end is past the end of
        // the buffer so we don't have to here
        getRandomValues(bytes.slice(generated, generated + MAX_BYTES))
      }
    } else {
      getRandomValues(bytes)
    }
  }
  return encoding.bufferToArray(bytes);
}

function bowserRandomBytes(length) {
  return window.crypto.getRandomValues(new Uint8Array(length))
}

function generateKey(length) {
  const _length = (length || 256) / 8;
  //const bytes = bowserRandomBytes(_length)
  const bytes = randomBytes(_length);
  const result = (0, utils.convertBufferToArrayBuffer)(encoding.arrayToBuffer(bytes));
  return result;
}

exports.generateKey = generateKey;
async function verifyHmac(payload, key) {
  const cipherText = encoding.hexToArray(payload.data);
  const iv = encoding.hexToArray(payload.iv);
  const hmac = encoding.hexToArray(payload.hmac);
  const hmacHex = encoding.arrayToHex(hmac, false);
  const unsigned = encoding.concatArrays(cipherText, iv);
  const chmac = await crypto.hmacSha256Sign(key, unsigned);
  const chmacHex = encoding.arrayToHex(chmac, false);
  if (encoding.removeHexPrefix(hmacHex) === encoding.removeHexPrefix(chmacHex)) {
    return true;
  }
  return false;
}

exports.verifyHmac = verifyHmac;
async function encrypt(data, key, providedIv) {
  const _key = encoding.bufferToArray((0, utils.convertArrayBufferToBuffer)(key));
  const ivArrayBuffer = providedIv || ( generateKey(128));
  const iv = encoding.bufferToArray((0, utils.convertArrayBufferToBuffer)(ivArrayBuffer));
  const ivHex = encoding.arrayToHex(iv, false);
  const contentString = JSON.stringify(data);
  const content = encoding.utf8ToArray(contentString);
  const cipherText = await crypto.aesCbcEncrypt(iv, _key, content);
  const cipherTextHex = encoding.arrayToHex(cipherText, false);
  const unsigned = encoding.concatArrays(cipherText, iv);
  const hmac =await  crypto.hmacSha256Sign(_key, unsigned);
  const hmacHex = encoding.arrayToHex(hmac, false);
  return {
    data: cipherTextHex,
    hmac: hmacHex,
    iv: ivHex,
  };

}

exports.encrypt = encrypt;
async function decrypt(payload, key) {
  const _key = encoding.bufferToArray((0, utils.convertArrayBufferToBuffer)(key));
  if (!_key) {
    throw new Error("Missing key: required for decryption");
  }
  const verified = await  verifyHmac(payload, _key);
  if (!verified) {
    return null;
  }
  const cipherText = encoding.hexToArray(payload.data);
  const iv = encoding.hexToArray(payload.iv);
  const buffer =await  crypto.aesCbcDecrypt(iv, _key, cipherText);
  const utf8 = encoding.arrayToUtf8(buffer);
  let data;
  try {
    data = JSON.parse(utf8);
  }
  catch (error) {
    return null;
  }
  return data;
}
exports.decrypt = decrypt;
